package org.devcake.groovy.autoimports

/**
 * Created by IntelliJ IDEA.
 * User: Admin
 * Date: 12.08.2010
 * Time: 9:43:11
 * To change this template use File | Settings | File Templates.
 */

import org.codehaus.groovy.control.SourceUnit
import org.codehaus.groovy.control.CompilationUnit
import org.codehaus.groovy.control.CompilerConfiguration
import org.codehaus.groovy.control.Phases
import org.codehaus.groovy.control.CompilationFailedException
import org.codehaus.groovy.ast.ImportNode
import org.codehaus.groovy.ast.ModuleNode
import org.codehaus.groovy.ast.ClassNode
import java.security.CodeSource
import org.codehaus.groovy.ast.PackageNode
import org.codehaus.groovy.control.CompilationUnit.SourceUnitOperation

/**
 * Helping class that allows to run selected part of script in context of whole script.
 *
 * @author Sergii Bondarenko
 *
 * Created on 2009-02-01
 */
public class AutoImportsGroovyShell extends GroovyShell {
    public AutoImportsGroovyShell() {
        this(null, new Binding())
    }

    public AutoImportsGroovyShell(Binding binding) {
        this(null, binding)
    }

    public AutoImportsGroovyShell(CompilerConfiguration config) {
        this(new Binding(), config)
    }

    public AutoImportsGroovyShell(Binding binding, CompilerConfiguration config) {
        this(null, binding, config)
    }

    public AutoImportsGroovyShell(ClassLoader parent, Binding binding) {
        this(parent, binding, CompilerConfiguration.DEFAULT)
    }

    public AutoImportsGroovyShell(ClassLoader parent) {
        this(parent, new Binding(), CompilerConfiguration.DEFAULT)
    }
    public AutoImportsGroovyShell(ClassLoader parent, Binding binding, final CompilerConfiguration config) {
        super(parent, binding, config)
        ClassLoader parentLoader = (parent!=null)?parent:GroovyShell.class.getClassLoader()
        GroovyShell.class.metaClass.setAttribute(this, 'loader', new AutoImportsGroovyClassLoader(parentLoader,config))
    }

    public Object runSelected(String fullScriptText, String selectedText, String fileName, List list) throws CompilationFailedException {
        def classLoader = (AutoImportsGroovyClassLoader) getClassLoader()
        try {
            try {
                def fullSource = new GroovyCodeSource(fullScriptText, 'FullScript', '/groovy.shell')
                def loader = new GroovyClassLoader()
                def compilationUnit = new CompilationUnit(CompilerConfiguration.DEFAULT, fullSource.getCodeSource(), loader)
                def su = compilationUnit.addSource('FullScript', new ByteArrayInputStream(fullScriptText.getBytes("UTF-8")))
                compilationUnit.compile(Phases.CONVERSION)
                classLoader.defaultImports = su.AST.imports
                classLoader.defaultImportPackages = su.AST.importPackages
                classLoader.defaultStaticImportClasses = su.AST.staticImportClasses
                classLoader.defaultStaticImportAliases = su.AST.staticImportAliases
                classLoader.defaultPackage = su.AST.package
                classLoader.targetFileName = fileName;
            } catch (CompilationFailedException e) {
                //e.printStackTrace(System.out)
                //since we are trying to run selected part of script, compilation errors in whole script are silently ignored
            }
            return run(selectedText, fileName, list as String[])
        } catch(CompilationFailedException e) {
            //e.printStackTrace(System.out)
            throw e
        } finally {
            classLoader.defaultImports = []
            classLoader.defaultImportPackages = []
            classLoader.defaultStaticImportClasses = [:]
            classLoader.defaultStaticImportAliases = [:]
            classLoader.targetFileName = null;
        }
    }

}

public static class AutoImportsGroovyClassLoader extends GroovyClassLoader {
    List<ImportNode> defaultImports = []
    List<String> defaultImportPackages = []
    Map<String, ClassNode> defaultStaticImportClasses = [:]
    Map<String, ClassNode> defaultStaticImportAliases = [:]
    PackageNode defaultPackage
    String targetFileName
    private boolean applyDefaultImports = false

    public AutoImportsGroovyClassLoader() {
        super()
    }
    public AutoImportsGroovyClassLoader(ClassLoader loader, CompilerConfiguration config) {
        super(loader, config)
    }
    public Class parseClass(GroovyCodeSource codeSource, boolean shouldCacheSource) throws CompilationFailedException {
        if (codeSource.getName() == targetFileName) {
            applyDefaultImports = true
        }
        try {
            return super.parseClass(codeSource, shouldCacheSource)
        } finally {
            applyDefaultImports = false
        }
    }
    protected CompilationUnit createCompilationUnit(CompilerConfiguration config, CodeSource source) {
        CompilationUnit cu = super.createCompilationUnit(config, source)
        if (applyDefaultImports) {
            def importsOperation = new AutoImportsOperation(
                loader:this,
                defaultImports:defaultImports,
                defaultImportPackages:defaultImportPackages,
                defaultStaticImportClasses:defaultStaticImportClasses,
                defaultStaticImportAliases:defaultStaticImportAliases)

            cu.addPhaseOperation(importsOperation, Phases.CONVERSION)
            cu.addPhaseOperation(new SourceUnitOperation() {
                void call(SourceUnit sourceUnit) {
                    if (defaultPackage?.name) {
                        if (!sourceUnit.getAST().packageName) {
                            sourceUnit.getAST().package = defaultPackage
                        }
                        sourceUnit.getAST().classes.each {
                            if (!it.packageName) {
                                it.name = defaultPackage.name + it.name
                            }
                        }
                    }
                }
            }, Phases.CONVERSION)
        }
        return cu
    }
    public boolean classIsAvailable(String className) {
        def end = false
        /* we are looking for specified class, and also check, maybe it is internal class
         * like java.util.Map$Entry or java.util$Map$Entry or ...
         */
        while(!end) {
            end = !className.contains('.')
            try {
                loadClass(className, false);
                return true
            } catch (ClassNoFoundException) {
            }
            className = className.replaceAll(/^(.*)\.([^\.]*)$/, /$1\$$2/)
        }
        return false
    }
}

private static class AutoImportsOperation extends CompilationUnit.SourceUnitOperation {
    AutoImportsGroovyClassLoader loader
    List<ImportNode> defaultImports = []
    List<String> defaultImportPackages = []
    Map<String, ClassNode> defaultStaticImportClasses = [:]
    Map<String, ClassNode> defaultStaticImportAliases = [:]

    public void call(SourceUnit source) throws CompilationFailedException {
        ModuleNode module = source.getAST()
        defaultImports.each{ if (module.getImport(it.alias) == null && loader.classIsAvailable(it.type.name)) module.addImport(it.alias, it.type) }
        defaultImportPackages.each{ module.addImportPackage(it) }
        defaultStaticImportClasses.each{ if (loader.classIsAvailable(it.value.name)) module.addStaticImportClass(it.key, it.value) }
        defaultStaticImportAliases.each{ if (!module.staticImportAliases.containsKey(it.key) && loader.classIsAvailable(it.value.name)) module.addStaticMethodOrField(it.value, it.key, it.key) }
    }
}